@@ -16,8 +16,8 @@
 *
 * Authors:
 *
-*  - Christopher <sahib> Pahl 2010-2015 (https://github.com/sahib)
-*  - Daniel <SeeSpotRun> T.   2014-2015 (https://github.com/SeeSpotRun)
+*  - Christopher <sahib> Pahl 2010-2017 (https://github.com/sahib)
+*  - Daniel <SeeSpotRun> T.   2014-2017 (https://github.com/SeeSpotRun)
 *
 * Hosted on http://github.com/sahib/rmlint
 **/
@@ -27,10 +27,25 @@
 
 #include <stdio.h>
 
-#include "pathtricia.h"
 #include "checksum.h"
+#include "pathtricia.h"
 #include "utilities.h"
 
+/* Struct for paths passed to rmlint from command line (or stdin) */
+typedef struct RmPath {
+    /* the RealPath of the passed string */
+    char *path;
+
+    /* index number (command line order) */
+    guint idx;
+
+    /* whether path was tagged as preferred path */
+    bool is_prefd;
+
+    /* whether to treat all files under path as one filesystem */
+    bool treat_as_single_vol;
+} RmPath;
+
 /* Storage struct for all options settable in cmdline. */
 typedef struct RmCfg {
     gboolean with_color;
@@ -60,6 +75,7 @@ typedef struct RmCfg {
     gboolean match_with_extension;
     gboolean match_without_extension;
     gboolean merge_directories;
+    gboolean honour_dir_layout;
     gboolean write_cksum_to_xattr;
     gboolean read_cksum_from_xattr;
     gboolean clear_xattr_fields;
@@ -69,6 +85,7 @@ typedef struct RmCfg {
     gboolean fake_fiemap;
     gboolean progress_enabled;
     gboolean list_mounts;
+    gboolean replay;
 
     int permissions;
 
@@ -82,13 +99,22 @@ typedef struct RmCfg {
 
     gboolean use_absolute_start_offset;
     gboolean use_absolute_end_offset;
+
     RmOff skip_start_offset;
     RmOff skip_end_offset;
 
-    /* all paths we should traverse (NULL terminated vector) */
-    char **paths;
-    char *is_prefd;
+    /* paths passed by command line */
+    GSList *paths;
+    GSList *json_paths;
+    guint path_count;
+
+    /* working dir rmlint called from */
     char *iwd;
+
+	/* Path to the rmlint binary of this run */
+	char *full_argv0_path;
+
+    /* the full command line */
     char *joined_argv;
 
     char *sort_criteria;
@@ -119,6 +145,12 @@ typedef struct RmCfg {
      * the end of the program run and printed then.
      */
     gboolean cache_file_structs;
+
+	/* Instead of running in duplicate detection mode,
+	 * check if the passed arguments are equal files
+	 * (or directories)
+	 */
+	gboolean run_equal_mode;
 } RmCfg;
 
 /**
@@ -126,4 +158,14 @@ typedef struct RmCfg {
  */
 void rm_cfg_set_default(RmCfg *cfg);
 
+/**
+ * @brief check and add a path to cfg->paths.
+ */
+guint rm_cfg_add_path(RmCfg *cfg, bool is_prefd, const char *path);
+
+/**
+ * @brief free all data associated with cfg->paths.
+ */
+void rm_cfg_free_paths(RmCfg *cfg);
+
 #endif /* end of include guard */
